# fireprompt/promptasync.py
# SPDX-License-Identifier: MIT

# imports
import inspect

from typing import Any
from typing import Callable
from typing import Optional
from jinja2 import Environment
from litellm import acompletion
from fireprompt.types import LLM
from fireprompt.prompt import FirePrompt


class FirePromptAsync(FirePrompt):

    def __init__(
        self,
        func: Callable,
        model: Optional[LLM] = None,
        on_complete: Optional[Callable] = None,
    ) -> None:
        """Initialize FirePromptAsync class."""
        super().__init__(
            func,
            model,
            on_complete,
            env=Environment(
                enable_async=True
            )
        )

        if on_complete:
            if not inspect.iscoroutinefunction(on_complete):
                raise ValueError(
                    "callback must be a coroutine function"
                )

    async def _render(self, *args: Any, **kwargs: Any) -> list[dict[str, str]]:
        """Render the prompt."""
        messages = []

        for m in self._parse_func_doc():
            messages.append({
                "role": m["role"],
                "content": await self._env.from_string(
                    m["content"]
                ).render_async(
                    **self._func.bind(
                        *args,
                        **kwargs
                    )
                )
            })

        return messages

    async def _call_llm(self, messages: list[dict[str, str]]) -> Any:
        """Async call LLM."""
        return self._parse_response(
            await acompletion(
                **self._get_model_config(),
                model=self._model.name,
                messages=messages,
                response_format=self._get_response_format()
            )
        )

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Async call to the prompt."""
        # call LLM
        response = await self._call_llm(
            await self._render(
                *args,
                **kwargs
            )
        )

        if self._on_complete:
            return await self._on_complete(response)

        return response



# fireprompt/api.py
# SPDX-License-Identifier: MIT

# imports
import inspect

from typing import Optional
from typing import Callable
from fireprompt.types import LLM
from fireprompt.prompt import FirePrompt
from fireprompt.promptasync import FirePromptAsync


def prompt(
    *,
    model: Optional[LLM] = None,
    on_complete: Optional[Callable] = None
) -> FirePrompt | FirePromptAsync:
    """Prompt decorator."""
    def decorator(func: Callable) -> Callable:
        """Inner decorator function."""
        if not inspect.iscoroutinefunction(func):
            return FirePrompt(
                func=func,
                model=model,
                on_complete=on_complete
            )

        return FirePromptAsync(
            func=func,
            model=model,
            on_complete=on_complete
        )

    return decorator



# fireprompt/types.py
# SPDX-License-Identifier: MIT

# imports
from typing import Any
from pydantic import Field
from pydantic import BaseModel


class LLMConfig(BaseModel):
    """LLM configuration."""
    max_tokens: int | None = Field(None, ge=0)
    temperature: float | None = Field(None, ge=0.0, le=2.0)
    top_p: float | None = Field(None, ge=0.0, le=1.0)
    extra: dict[str, Any] | None = Field(None, description="provider-specific")


class LLMConfigPreset:
    """Pre-configured LLM settings for common use cases."""
    DEFAULT = LLMConfig(temperature=0.7)
    CREATIVE = LLMConfig(temperature=0.9, top_p=0.95)
    PRECISE = LLMConfig(temperature=0.3, top_p=0.9)
    DETERMINISTIC = LLMConfig(temperature=0.0, top_p=1.0)


class LLM(BaseModel):
    """LLM model."""
    name: str
    config: LLMConfig = Field(default_factory=lambda: LLMConfigPreset.DEFAULT)



# fireprompt/prompt.py
# SPDX-License-Identifier: MIT

# imports
import yaml
import inspect

from typing import Any
from typing import Callable
from typing import Optional
from pydantic import BaseModel
from litellm import completion
from jinja2 import Environment
from fireprompt.types import LLM
from fireprompt.logger import Logger
from fireprompt.func import FuncWrapper


class FirePrompt:

    def __init__(
        self,
        func: Callable,
        model: Optional[LLM] = None,
        on_complete: Optional[Callable] = None,
        env: Optional[Environment] = None,
    ) -> None:
        """Initialize FirePrompt class."""
        if not model:
            raise ValueError("model is required")

        self._func = FuncWrapper(func)
        self._logger = Logger.get(self.__class__.__name__)
        self._on_complete = on_complete
        self._model = model
        self._env = env or Environment()

    def _parse_func_doc(self) -> dict[str, Any]:
        """Parse function docstring."""
        doc = self._func.doc

        if not doc:
            raise ValueError(
                f"function {self._func.__name__} must have a docstring"
            )

        try:
            parsed = yaml.safe_load(doc)  # load yaml

            if not isinstance(parsed, list):
                # if not a list, raise an error
                raise ValueError("invalid yaml docstring")

            return parsed
        except yaml.YAMLError as e:
            # if not valid yaml, raise an error
            raise ValueError(
                f"invalid yaml docstring: {e}"
            ) from e

    def _get_schema(self) -> dict[str, Any]:
        """Get schema."""
        schema = self._func.return_annotation.model_json_schema()

        schema.update({
            # disallow additional properties (strict mode)
            "additionalProperties": False
        })

        return schema

    def _get_response_format(self) -> dict[str, Any]:
        """Get response format."""
        if not hasattr(
            self._func.return_annotation,
            "model_json_schema"
        ):
            # handle non-Pydantic return types
            return None

        return {
            "type": "json_schema",
            "json_schema": {
                "schema": self._get_schema(),
                "name": self._func.return_annotation.__name__,
                "strict": True
            }
        }

    def _render(self, *args: Any, **kwargs: Any) -> list[dict[str, str]]:
        """Render the prompt."""
        messages = []

        for m in self._parse_func_doc():
            messages.append({
                "role": m["role"],
                "content": self._env.from_string(
                    m["content"]
                ).render(
                    **self._func.bind(
                        *args,
                        **kwargs
                    )
                )
            })

        return messages

    def _get_model_config(self) -> dict[str, Any]:
        """Get model config and merge extra params."""
        if not self._model.config:
            return {}

        config = self._model.config.model_dump(exclude_none=True)
        extra = config.pop("extra", {}) or {}
        return {**config, **extra}

    def _parse_response(self, response: Any) -> Any:
        """Parse response."""
        # extract content from response
        content = response.choices[0].message.content

        if (
            self._func.return_annotation
            and self._func.return_annotation != inspect.Signature.empty
            and isinstance(self._func.return_annotation, type)
            and issubclass(self._func.return_annotation, BaseModel)
        ):
            return self._func.return_annotation.model_validate_json(content)

        # return raw response
        return content

    def _call_llm(self, messages: list[dict[str, str]]) -> Any:
        """Call LLM."""
        return self._parse_response(
            completion(
                **self._get_model_config(),
                model=self._model.name,
                messages=messages,
                response_format=self._get_response_format()
            )
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Sync call to the prompt."""
        # call LLM
        response = self._call_llm(
            self._render(
                *args,
                **kwargs
            )
        )

        if self._on_complete:
            # if a callback is provided, call it with the response
            # return the result of the callback
            return self._on_complete(response)

        return response



# fireprompt/__init__.py
# SPDX-License-Identifier: MIT

# imports
import warnings

from fireprompt.types import LLM
from fireprompt.api import prompt
from fireprompt.types import LLMConfig
from fireprompt.types import LLMConfigPreset
from fireprompt.logger import set_debug_mode

# suppress litellm cleanup warnings
warnings.filterwarnings(
    "ignore",
    category=RuntimeWarning,
    message=".*coroutine.*was never awaited.*"
)


# exports public APIs
__all__ = [
    "prompt",
    "LLM",
    "LLMConfig",
    "LLMConfigPreset",
    "enable_logging",
    "disable_logging"
]

# version
__version__ = "0.1.2"


# logging
def enable_logging() -> None:
    """Enable logging."""
    set_debug_mode(True)


def disable_logging() -> None:
    """Disable logging."""
    set_debug_mode(False)



# fireprompt/func.py
# SPDX-License-Identifier: MIT

# imports
import inspect
import functools

from typing import Any
from typing import Callable
from fireprompt.logger import Logger


class FuncWrapper:

    def __init__(self, func: Callable) -> None:
        """FuncWrapper class."""
        self._func = func
        self._logger = Logger.get(self.__class__.__name__)

        # preserve original function metadata
        functools.update_wrapper(self, func)

    @functools.cached_property
    def signature(self) -> inspect.Signature:
        """Get signature of original function."""
        return inspect.signature(self._func)

    @property
    def return_annotation(self) -> Any:
        """Get return annotation of original function."""
        return self.signature.return_annotation

    @property
    def doc(self) -> str | None:
        """Get docstring of original function."""
        return inspect.getdoc(self._func)

    def bind(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Bind arguments and apply defaults."""
        bound = self.signature.bind(*args, **kwargs)
        bound.apply_defaults()

        return bound.arguments



# fireprompt/logger.py
# SPDX-License-Identifier: MIT

# imports
import logging


# global debug mode state
_debug_mode = False

# track our loggers
_fireprompt_loggers = set()


class Logger:

    @staticmethod
    def get(name: str = "fireprompt"):
        """Get Logger instance."""
        ref = logging.getLogger(name)
        _fireprompt_loggers.add(name)

        if not ref.handlers:
            handler = logging.StreamHandler()

            # set formatter
            handler.setFormatter(
                logging.Formatter(
                    "[%(asctime)s] [%(name)s] [%(levelname)s]: %(message)s",
                    datefmt="%H:%M:%S"
                )
            )

            ref.addHandler(handler)
            ref.setLevel(logging.DEBUG if _debug_mode else logging.WARNING)

        return ref


def set_debug_mode(enabled: bool) -> None:
    """Set debug mode."""
    global _debug_mode
    _debug_mode = enabled

    for name in _fireprompt_loggers:
        logger = logging.getLogger(name)
        logger.setLevel(
            logging.DEBUG if enabled else logging.WARNING
        )
